import hashlib
from json import JSONDecoder, JSONEncoder
from pathlib import Path

from menagerie.Items.AbstractItem import AbstractItem

__all__ = ('CacheHandler',)

from menagerie.Settings import Settings


class CacheHandler:
    HASH_ALGORITHM = 'sha256'

    def __init__(self, folder: Path, settings: Settings):
        self.enabled = settings['cache_enabled']
        self.cache_folder = folder
        self.cache_file = Path(self.cache_folder, "item_cache.json")
        self.decoder = JSONDecoder()
        self.encoder = JSONEncoder()
        self.cache_folder.mkdir(parents=True, exist_ok=True)
        if self.cache_file.exists() is False:
            self.cache_file.write_text("{}", encoding='utf-8')
        self.cache_data = self.decoder.decode(self.cache_file.read_text(encoding='utf-8'))

    def get_hash(self, content: bytes) -> str:
        hasher = hashlib.new(self.HASH_ALGORITHM)
        hasher.update(content)
        return hasher.hexdigest()

    def check_item_changed(self, item: AbstractItem) -> bool:
        if self.enabled is False:
            return True
        cache_key = str(item.in_path)
        hashed = self.get_hash(item.get_path_to_open().read_bytes())
        return self.cache_data.get(cache_key, "") != hashed

    def add_item(self, item: AbstractItem) -> None:
        if self.enabled is False:
            return
        cache_key = str(item.in_path)
        hashed = self.get_hash(item.get_path_to_open().read_bytes())
        self.cache_data[cache_key] = hashed
        self.cache_file.write_text(self.encoder.encode(self.cache_data))



import importlib.resources as pkg_resources
from json import load, loads
from json.decoder import JSONDecodeError
from os import getenv
from pathlib import Path

from jsonschema import validate
from jsonschema.exceptions import ValidationError
from ndicts.ndicts import NestedDict

from menagerie import schemas
from menagerie.utils.logger import Logger

__all__ = (
    'Settings',
    'setup_settings'
)

Settings = NestedDict({
    'config_path': None,
    'out_dir': None,
    'content_dir': None,
    'url_prefix': None,
    'default-toc': True,
    'base_url': '',
    'cache_enabled': True,
    'log_level': "Info",
    'themes': {
        'bootstrap': "https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css",
        'highlight_js': "https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.4.0/styles/default.min.css",
        'theme': "light"
    },
    'brand': {
        'app_name': "My App",
        'short_name': "My App",
        'favicon_folder': 'images/fav/',
        'navbar_icon': None,
        'navbar_icon_size': (100, 100),
        'socials': [],
        'show_gen_date': True,
        'footer': {
            'show_gen_time': True,
            'show_back_to_top': True,
            'show_made_with': True,
            'links': []
        },
        'meta': {
            'description': "",
            'keywords': [],
            'categories': [],
            'image': "",
            'image_alt': "Logo",
            'theme_color': "#333333",
            'bg_color': "#ffffff"
        }
    },
    'styles': {
        'base': None,
        'schema': None
    },
    'paths': {
        'out': "out/",
        'content': 'content/',
        'pages': 'pages/',
        'static': 'static/',
        'cache': '.m_cache/'
    },
    'minify':
        {
            'html': True,
            'css': True,
            'js': True,
            'xml': True,
            'json': True
        }
})


def setup_settings(config_path: Path):
    """
        Loads settings from a given config file and validates it against the config schema

        :param config_path: The path to the config file to load
        :type config_path: Path 
    """

    schema = loads(pkg_resources.read_text(schemas, 'config_schema.json'))

    try:
        config = load(config_path.open(mode='r', encoding='utf-8'))
        validate(config, schema)
        Settings.update(NestedDict(config))
        Settings['out_dir'] = Path(Settings['paths', 'out'])
        Settings['content_dir'] = Path(Settings['paths', 'content'])
        Settings['url_prefix'] = getenv("URL_PREFIX", "")
        Settings['config_path'] = config_path
        Logger.update_level_from_string(Settings['log_level'])
    except FileNotFoundError:
        Logger.log_error(f"Can't find config file: `{config_path.as_posix()}`")
    except JSONDecodeError as e:
        Logger.log_error(f"Error parsing config: {e}")
    except ValidationError as e:
        Logger.log_error(f"Invalid config: {e.message}")



import sys
from pathlib import Path

from menagerie.Items.Managers.MetaManager import MetaManager
from menagerie.Items.Managers.PageManager import PageManager
from menagerie.Items.Managers.StaticManager import StaticManager
from menagerie.Settings import setup_settings, Settings
from menagerie.SiteGen import SiteGen
from menagerie.utils.logger import Logger

__all__ = (
    'generate',
)


def generate(*argv):
    config_path = None
    for arg in argv:
        if '--config=' in arg:
            try:
                config_path = Path(arg.split('=')[1])
                break
            except IndexError:
                Logger.log_error("Usage: python generate.py --config=CONFIG_PATH")
                return
    if config_path is None:
        config_path = Path('config.json')
    Logger.log_info("Loading Config")
    setup_settings(config_path)
    managers = (PageManager, StaticManager, MetaManager)
    Logger.log_info("Setting Up")
    gen = SiteGen(Settings, managers)
    Logger.log_info("Beginning Generation")
    Logger.log_info("Finding Content")
    gen.find()
    Logger.log_info("Initializing Meta Data")
    gen.initialize()
    Logger.log_info("Generating Content")
    gen.generate()
    Logger.log_info("Done!")


if __name__ == "__main__":
    generate(*sys.argv)



from menagerie.generate import generate
from menagerie.new_project import new_project

__version__ = '0.0.9'
__help__ = """
                Usage:
                    menagerie new-project [name] : Creates a new menagerie project with a given name
                    menagerie generate [--config=...] : Generates a site with the given config
                    menagerie help : Shows this message
                    menagerie version : Shows the version
            """



"""
    A set of functions that are helpful when outputting data
"""
from __future__ import annotations

import sys
from dataclasses import dataclass
from enum import IntEnum

__all__ = (
    'ConsoleColorMode',
    'COLOR_MODE',
    'ConsoleTextStyle',
    'FourBitConsoleColors',
    'color_text',
    'reset_format',
    'progressbar'
)


class ConsoleColorMode(IntEnum):
    FOUR_BIT = 0
    EIGHT_BIT = 1
    RGB = 2


COLOR_MODE = ConsoleColorMode.FOUR_BIT


class FourBitConsoleColors(IntEnum):
    """
        Use this static class to select colors in :class:ConsoleTextStyle
    """

    BLACK = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    PURPLE = 5
    CYAN = 6
    WHITE = 7


@dataclass
class ConsoleTextStyle:
    """
        A set of options for outputting colored and bolded text to the console
    """

    fg_color: int | tuple[int] = None
    bg_color: int | tuple[int] = None
    bold: bool = False
    underline: bool = False
    overline: bool = False
    blink: bool = False
    high_intensity: bool = False

    def get_flags(self):
        return self.bold, self.underline, self.overline, self.blink


flag_codes = (
    (1, None),
    (4, 24),
    (53, 55),
    (5, 25),
)


# \033[color;bg;flagsmMESSAGE\033[0m


def handle_flags(style):
    flags = style.get_flags()
    flag_string = ""
    for flag, possible in enumerate(flag_codes):
        if flags[flag] is True and possible[0] is not None:
            flag_string += f"{possible[0]};"
        elif flags[flag] is False and possible[1] is not None:
            flag_string += f"{possible[1]};"
    return flag_string[:-1]


def handle_color(color_type, style: ConsoleTextStyle, mode):
    color = style.fg_color if color_type == 0 else style.bg_color
    if color is None:
        return ""
    else:
        if mode == ConsoleColorMode.RGB:
            return str(38 if color_type == 0 else 48) + f";2;{color[0]};{color[1]};{color[2]};"
        elif mode == ConsoleColorMode.FOUR_BIT:
            return str((10 if color_type == 1 else 0) + (60 if style.high_intensity else 0) + int(color) + 30) + ";"
        elif mode == ConsoleColorMode.EIGHT_BIT:
            return str(38 if color_type == 0 else 48) + ';5;' + str(color) + ";"


def color_text(in_str: str, style: ConsoleTextStyle, reset: bool = True,
               override_color_mode: ConsoleColorMode = None) -> str:
    """
        Colors the given string according to the provided :py:class:`ConsoleTextStyle`

        :param reset: Whether to reset the formatting after the string
        :type reset: bool
        :param override_color_mode: Overrides the COLOR_MODE global option
        :type override_color_mode: ConsoleColorMode
        :param in_str: The string to color
        :type in_str: str
        :param style: The style to apply
        :type style: ConsoleTextStyle
        :returns: The colored version of the given text
        :rtype: str
    """

    if style is None:
        return in_str
    else:
        if override_color_mode is None:
            color_mode = COLOR_MODE
        else:
            color_mode = override_color_mode
        start_sequence = "\033["
        fg_color_sequence = handle_color(0, style, color_mode)
        bg_color_sequence = handle_color(1, style, color_mode)
        flag_sequence = handle_flags(style)
        end_sequence = '\033[0m' if reset else ""
        return start_sequence + fg_color_sequence + bg_color_sequence + flag_sequence + 'm' + in_str + end_sequence


def reset_format():
    print('\033[0m', end='\r')


def progressbar(it, prefix="", size=60):
    """
        Wraps a given iterator or generator (use a cast to list) to print a progress bar to the console

        :param it: The iterator to apply the bar to
        :type it: iterable
        :param prefix: Characters to show before the bar
        :type prefix: str
        :param size: The size of the bar
        :type size: int
    """

    count = len(it)

    def show(j):
        x = int(size * j / count)
        sys.stdout.write("%s[%s%s] %i/%i\r" % (prefix, "#" * x, "." * (size - x), j, count))
        sys.stdout.flush()

    show(0)
    for i, item in enumerate(it):
        yield item
        show(i + 1)
    sys.stdout.write("\n")
    sys.stdout.flush()



import importlib.resources as pkg_resources
from json import loads, load, JSONDecodeError
from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate
from markdown import Markdown

from menagerie import schemas
from menagerie.Items.Managers.AbstractManager import AbstractManager
from menagerie.Items.Pages.AbstractPage import AbstractPage
from menagerie.utils.str_util import pretty_title
from menagerie.Items.Pages.HTMLPage import HTMLPage
from menagerie.Items.Pages.MDPage import MDPage
from menagerie.Items.Pages.Schema.JSONSchema import JSONSchema
from menagerie.Items.Pages.Schema.XMLSchema import XMLSchema
from menagerie.utils.logger import Logger


class NavItem:

    def __init__(self, content: str | list['NavItem'], is_folder: bool, meta: dict[str, object]):
        self.is_folder = is_folder
        self.titles = []
        self.meta = meta
        if is_folder:
            self.content = []
        else:
            self.content = content

    def add_item(self, nav_item: 'NavItem') -> None:
        if self.is_folder:
            self.content.append(nav_item)

    def iter_items(self) -> None:
        for page in sorted(self.content, key=lambda p: p.meta['sort_priority'], reverse=True):
            yield page

    def setup_titles(self) -> None:
        if self.is_folder:
            self.titles = [str(p.meta['title']).lower() for p in self.content]


class PageManager(AbstractManager):
    root_dir = 'pages'
    item_types: AbstractPage = (MDPage, HTMLPage, XMLSchema, JSONSchema)
    context: dict[str, object] = {

    }

    env: Environment = None
    router: dict[str, str] = {}

    def __init__(self, site_info) -> None:
        super(PageManager, self).__init__(site_info)
        self.env = Environment(loader=FileSystemLoader(self.gen.settings['content_dir']))
        self.root_dir = Path(self.gen.settings['paths', 'pages'])
        self.group_schema = loads(pkg_resources.read_text(schemas, 'folder_schema.json', encoding='utf-8'))

    def find(self):
        super(PageManager, self).find()

    def initialize(self):
        for item in self.items:
            item.initialize()
        self.router = {p.meta['title'].lower(): str(p.out_path.relative_to(self.gen.settings['out_dir']).as_posix()) for
                       p in self.items}
        self.gen.shared_info['pages'] = filter(lambda i: i.meta['hide_in_nav'] is False, self.items)
        self.filter_md = Markdown(extensions=['extra'], output_format='html5')

    def route(self, title: str) -> str:
        return self.gen.settings['url_prefix'] + self.router.get(title.lower(), "#")

    def make_nav_item_from_page(self, page: AbstractPage) -> NavItem:
        return NavItem(self.route(page.meta['title']), False, page.meta)

    def generate_nav_items(self) -> list[NavItem]:
        root_pages = []
        groups: dict[str, NavItem] = {}
        for page in filter(lambda p: p.meta['hide_in_nav'] is not True, self.items):
            if page.in_path.parent.name != '':
                if page.in_path.parent.name in groups.keys():
                    groups[page.in_path.parent.name].add_item(self.make_nav_item_from_page(page))
                else:
                    group_config_path = Path(page.get_path_to_open().parent, '_folder.json')
                    group_meta = {
                        'title': pretty_title(group_config_path.parent.name),
                        'sort_priority': 30
                    }
                    if group_config_path.exists():
                        try:
                            group_config = loads(group_config_path.read_text(encoding='utf-8'))
                            validate(group_config, self.group_schema)
                            group_meta.update(group_config)
                        except JSONDecodeError as jde:
                            Logger.log_error(f"Couldn't parse config for {group_config_path.name} : {str(jde)}")
                        except ValidationError as ve:
                            Logger.log_error(f"Invalid folder config for {group_config_path.name} : {str(ve)}")
                    new_group = NavItem([], True, group_meta)
                    new_group.add_item(self.make_nav_item_from_page(page))
                    groups[group_config_path.parent.name] = new_group
            else:
                new_page = self.make_nav_item_from_page(page)
                root_pages.append(new_page)
        for group in groups.values():
            group.setup_titles()
        return sorted(list(groups.values()) + root_pages, key=lambda n: n.meta['sort_priority'], reverse=True)

    def generate(self):
        self.gen.settings['out_dir'].mkdir(parents=True, exist_ok=True)
        filters = {
            'route': self.route,
            'static': self.gen.shared_info['static_filter'],
            'upper_first': lambda x: x[0].upper() + x[1:],
            'simple_md': lambda md: self.filter_md.convert(md)
        }
        da_globals = {
            'nav_items': self.generate_nav_items(),
        }
        self.base_env.filters.update(filters)
        self.base_env.globals.update(da_globals)
        self.env.filters.update(self.base_env.filters)
        self.env.globals.update(self.base_env.globals)
        for item in self.items:
            item.generate()
